# Main script for Taylor-Maccoll solver.
# VERSION: 1
# AUTHOR: Vojtech Pezlar
# LAST UPDATE: 07/04/2024
# EMAIL: pezlavoj@student.cvut.cz

#-----------------------------------------------------------------------------------------------------------------------------#
# Code Purpose:                                                                                                               #
# The following code solves the Taylor Maccoll (TM) equations for supersonic flow over a cone using shooting methods, given   #
# freestream conditions and cone angle. The solution is obtained by marching from the shockwave towards the cone,             #
# where stopping criteria are set such that the angular velocity at the surface of the cone is 0. The velocity components     #
# just after the shockwave, necessary as inputs for the TM equations are obtained using oblique shock relations. The angle at #
# which the angular velocity is zero is then compared with the assigned cone angle and if they differ,                        #
# the shockwave angle is then adjusted using the bi-section technique. When a correct shockwave angle is found,               #
# the corresponding solution of the TM flow is then distributed along rays (lines of constant angle).                         #
#                                                                                                                             #
# Reference: J.D. Anderson - Modern Compressible Flow With Historical Perspective                                             #
#-----------------------------------------------------------------------------------------------------------------------------#


# -----------------------------------------------#
#                    Libraries                   #
# -----------------------------------------------#

# 3rd party libraries
import numpy as np                     # General array and matrix manipulation package
from matplotlib import pyplot as plt   # Visualisation package

# Subroutines
import RK4                             # Runge-Kutta marching routine
import ODE_TM                          # Definition of Taylor-Maccoll equations
import ObliqueShock                    # Shock relations for values immediately after shock
from BiSection import BiSection        # Root finding routine
import Output                          # Output is formatted for Paraview software and stored as VTK file


# -----------------------------------------------#
#                    Freestream                  #
# -----------------------------------------------#
# Problem Settings
gamma      = 1.4  # Gas constant

cone_angle = np.deg2rad(10)  # Cone half-angle (Provide into brackets in degrees)
M_inf      = 2               # Free-stream Mach
T_inf      = 100             # Freestream temperature [K]
P_inf      = 125000          # Freestream Pressure [Pa]
R_gas      = 287.058         # Gas Constant [SI]


# -----------------------------------------------#
#                 Solver Settings                #
# -----------------------------------------------#
# Solver settings
delta       = -0.001  # Step (-(ve). bcs. angle is decreasing from shock-angle to cone-angle)
Criteria    = 1e-9  # Stopping Criteria for marching routine
tol         = 1e-8  # Tolerance for Root-Finding method

# Obtaining bounds for Bi-Section method {Tested values}
# If a desired combination of Mach and cone angle outputs error:
# The scalars a and b do not bound a root, widen the interval set by [Shock_min, Shock_max]
Shock_max   = np.deg2rad(70)
Shock_min   = np.arcsin((1 / M_inf) - (1 / M_inf) * 0.4)


# -----------------------------------------------#
#               Fittness Function                #
# -----------------------------------------------#

# Defining Fitness function controlling the Root-Finding routine (Bi-Section)

def Fitness(Shock_angle, M, GAMMA, cone):
    # Obtaining V_r and V_theta velocity components immediately after shock
    v = ObliqueShock.pastShock(Shock_angle, M, GAMMA)
    # Marching from given shock angle till V_theta == 0 within given stopping criteria
    q, angle = RK4.RK4_TM(ODE_TM.ODE_TM, Shock_angle, Criteria, delta, [v[0], v[1]], 'True', GAMMA)

    # Difference between final angle obtained from marching and prescribed cone_angle
    return angle[-1] - cone


# -----------------------------------------------#
#                   Shock Angle                  #
# -----------------------------------------------#

# Shock Angle (Rad) corresponding to the provided cone-angle and Free-stream Mach
Shock_Angle = BiSection(Fitness, Shock_max, Shock_min, M_inf, tol, 0, gamma, cone_angle)

# Shock Angle (Deg)
Shock_Angle_Deg = np.rad2deg(Shock_Angle)


# -----------------------------------------------#
#                    Flowfield                   #
# -----------------------------------------------#
# Shock angle returned by shooting method is now used to obtain the full flowfield

# Full Flow-field (All Rays)
v_final = ObliqueShock.pastShock(Shock_Angle, M_inf, gamma)
Flowfield, Angles = RK4.RK4_TM(ODE_TM.ODE_TM, Shock_Angle, Criteria, delta, [v_final[0], v_final[1]], 'True', gamma)

# Allocating Variable
M_Rays = np.zeros(len(Flowfield[:, 1]))

# Obtaining Mach at each ray, based on V_r and V_theta
for i in range(len(Flowfield[:, 1])):

    # Velocity = (V_r^2 + V_theta^2)^0.5
    VEL = np.sqrt(Flowfield[i, 0] ** 2 + Flowfield[i, 1] ** 2)

    # Mach -> Ref. J. Anderson, Modern Compressible Flow with Historical Perspective 3rd Ed.
    # Formula relating Mach number and velocity
    M_Rays[i] = np.sqrt(2 / ((gamma - 1) * (1 / (VEL ** 2) - 1)))

M_c = M_Rays[-1]  # Mach value at the surface of the cone


# Oblique Shock Relations
# Pressure ratio - Behind-Shock/Freestream
P2_Pinf = 1 + (2*gamma / (gamma + 1)) * ((M_inf * np.sin(Shock_Angle))**2 - 1)
# Pressure ratio - Cone-Surface/Freestream
Pc_Pinf = ((1 + ((gamma - 1)/2) * (v_final[4])**2) / (1 + ((gamma - 1)/2) * M_c**2))**(gamma / (gamma - 1)) * \
          (1 + (2*gamma/(gamma + 1))*((M_inf * np.sin(Shock_Angle))**2 - 1))

# Temperature ratio - Behind-Shock/Freestream
T2_Tinf = P2_Pinf * (2 + (gamma-1)*(M_inf * np.sin(Shock_Angle))**2)/((gamma+1)*(M_inf*np.sin(Shock_Angle))**2)
# Temperature ratio - Cone-Surface/Freestream
Tc_Tinf = ((1 + ((gamma - 1)/2) * (v_final[4])**2) / (1 + ((gamma - 1)/2) * M_c**2)) * \
          (1 + (2*gamma/(gamma + 1))*((M_inf * np.sin(Shock_Angle))**2 - 1)) * \
          (2 + (gamma-1)*(M_inf * np.sin(Shock_Angle))**2)/((gamma+1)*(M_inf*np.sin(Shock_Angle))**2)

# Density values calculated using e.o.s.
Rho_inf = P_inf / (R_gas * T_inf)
Rho_c = Pc_Pinf / (R_gas * Tc_Tinf)


# -----------------------------------------------#
#                   Visualization                #
# -----------------------------------------------#

N      = 100                    # Number of nodes along r
r      = np.linspace(0, 10, N)  # r - vector of values
phi    = np.flip(Angles)        # vector of angles (flipped, bcs. original it is in decreasing order)
M_Rays = np.flip(M_Rays)        # vector of Mach values along rays (-||-)
u_rad  = Flowfield[:, 0]
u_rad  = np.flip(u_rad)
u_phi  = Flowfield[:, 1]
u_phi  = np.flip(u_phi)

# Allocation
Mach   = np.zeros((len(phi), len(r)))
U_rad  = np.zeros((len(phi), len(r)))
U_phi  = np.zeros((len(phi), len(r)))
U_x    = np.zeros((len(phi), len(r)))
U_y    = np.zeros((len(phi), len(r)))
x      = np.zeros((len(phi), len(r)))
y      = np.zeros((len(phi), len(r)))

# Making matrices for visualization and changing F.O.R from Polar to Cart.
for i in range(len(phi)):
    Mach[i, :]    = M_Rays[i]
    U_rad[i, :]   = u_rad[i]
    U_phi[i, :]   = u_phi[i]
    for j in range(len(r)):
        x[i, j]   = np.cos(phi[i]) * r[j]
        y[i, j]   = np.sin(phi[i]) * r[j]
        U_x[i, j] = U_rad[i, j] * np.cos(phi[i]) - U_phi[i, j] * np.sin(phi[i])
        U_y[i, j] = U_rad[i, j] * np.sin(phi[i]) + U_phi[i, j] * np.cos(phi[i])

# Routine to convert data to paraview compatible VTK format
Output.Paraview(x, y, r, phi, U_x, U_y, U_rad, U_phi, Mach)


# -----------------------------------------------#
#                Printing Results                #
# -----------------------------------------------#

print("+-----------------------------------------------------------------+")
print("|                      FREESTREAM VALUES                          |")
print("+-----------------------------------------------------------------+")
print(" Free-stream Mach number                  = " + str("% .3f" % M_inf))
print(" Cone Angle [Deg]                         = " + str("% .3f" % np.rad2deg(cone_angle)))
print(" Free-stream Temperature [K]              = " + str("% .3f" % T_inf))
print(" Free-stream Pressure [Pa]                = " + str("% .3f" % P_inf))
print(" Free-stream Density [kg/m^3]             = " + str("% .3f" % Rho_inf))
print("+-----------------------------------------------------------------+")
print("|                           RESULTS                               |")
print("+-----------------------------------------------------------------+")
print(" Shock angle [Deg]                        = " + str("% .3f" % Shock_Angle_Deg))
print("+-----------------------------------------------------------------+")
print("|                            RATIOS                               |")
print("+-----------------------------------------------------------------+")
print(" T_1 / T_0 RATIO                          = " + str("% .3f" % T2_Tinf))
print(" P_1 / P_0 RATIO                          = " + str("% .3f" % P2_Pinf))
print(" T_c / T_0 RATIO                          = " + str("% .3f" % Tc_Tinf))
print(" P_c / P_0 RATIO                          = " + str("% .3f" % Pc_Pinf))
print("+-----------------------------------------------------------------+")
print("|                        SURFACES VALUES                          |")
print("+-----------------------------------------------------------------+")
print(" Cone Surface Mach number                 = " + str("% .3f" % M_c))
print(" Cone Surface Temperature [K]             = " + str("% .3f" % (T_inf * Tc_Tinf)))
print(" Cone Surface Pressure [Pa]               = " + str("% .3f" % (P_inf * Pc_Pinf)))
print(" Cone Surface Density [kg/m^3]            = " + str("% .3f" % Rho_c))
